#######################################
#PreMass
#genetic data format rearrangement for MASS package processing
#version:  1.0
#######################################

rm(list=ls(all=TRUE))
library(Matrix)
library(MetaSKAT)

#define commonly used constants
PACE = 1000000

##########################################
#preMASSscriptTemplateMake
#
##########################################
templateMaker = function(tempFileName = "preMassScriptTemplate.txt"){
stringX = "## THE FIRST STUDY : using RAREMETAL##
SOFTWARE = RAREMETAL
FILE_SCORE = <path>/study1.score.txt
FILE_COV = <path>/study1.cov.txt
FILE_GROUP = <path>/study1.group.txt

## THE FIRST STUDY : using MetaSKAT##
SOFTWARE = MetaSKAT
FILE_MSSD = <path>/study2.MSSD
FILE_MInfo = <path>/study2.Minfo

## THE FIRST STUDY : using seqMeta##
SOFTWARE = seqMeta
FILE_RDATA = <path>/study3.RData
FILE_OBJ = study3"

    write(stringX, tempFileName)
}

##########################################
#.rareMetalToMass.  Preprocess metaData from RareMetal program
#input files:
#	ifile1: score file name (full path)
#	ifile2: covariance file name (full path)
#	ifile3: group file name (full path)
#     ofile1: output meta data file name (full path)
#output:
##########################################
.rareMetalToMass = function(ifile1, ifile2, ifile3, ofile1){
    #pre-processing score file
    score = read.table(ifile1, header = FALSE, sep = "\t", comment.char = "#", as.is = T)[c(1,2,5,8,14,15)]
    scoreScan = scan(ifile1, sep = "\t", what = character(), quiet = T)
    samplesVal  = grep("##Samples", scoreScan, value = T) 
    sampleNo = as.numeric(unlist(strsplit(samplesVal, "="))[2])

    #pre-processing group file
    grp = scan(ifile3, sep = "", what = character(),comment.char = "#", quiet = T)
    grp = gsub(":A:T", "", grp)
    geneloc = grep("gene", grp)
    grpTab = NULL
    for ( i in 1:length(geneloc)){
        startpos = geneloc[i]+1
        if(i == length(geneloc)){
            endpos = length(grp)
        }else{
            endpos =  geneloc[i+1] - 1
        }
        tempTab = data.frame(snp_id = grp[startpos:endpos], stringsAsFactors = F)
        tempTab$gene_id = grp[geneloc[i]]
        grpTab = rbind(grpTab, tempTab)
    }

    #read cov file and repair cov table
    cov = read.table(ifile2, header = FALSE, sep = "\t", comment.char = "#", as.is = T)
    cov$snp = paste(cov$V1, cov$V2, sep = ":")
    covNew = NULL
    for(i in 1:dim(cov)[1]){
        temp = cov[i,]
        snpA = temp$snp
        snpB = paste(temp$V1,unlist(strsplit(temp$V3, ",")), sep = ":")
        val = unlist(strsplit(temp$V4, ","))
        geneA = grpTab$gene_id[temp$snp == grpTab$snp_id]
        snpBList = grpTab$snp_id[grpTab$gene_id == geneA]

        val = val[snpB %in% snpBList]
        snpB = snpB[snpB %in% snpBList]
        tempDf = data.frame(snpB = snpB, val = val, stringsAsFactors =F)
        tempDf$snpA = snpA
        tempDf$gene = geneA
        covNew = rbind(covNew, tempDf)
    }
    covNewFirst = covNew[covNew$snpA == covNew$snpB,c("snpA", "val"),]

    #processing.  two steps
    #step1
    rstOutVal = c("gene_id","snp_id","mac","N_obs","score_stat")
    rst = data.frame(snp_id = paste(score$V1, score$V2, sep = ":"),
                 mac = score$V8, 
                 N_obs = sampleNo,
                 v14 = score$V14,
                 v15 = score$V15,
                 score_ = score$V14/score$V15*(sampleNo)^0.5, stringsAsFactors = F)
    rst$orgSeq = 1:dim(rst)[1]
    rst = merge(rst, covNewFirst, by.x = "snp_id", by.y = "snpA", all.x = T, al.y = F)
    rst$score_stat = rst$score_*(as.numeric(rst$val))^0.5
    sigma = rst$v15[1]/(rst$N_obs[1] * as.numeric(rst$val[1]))^0.5; #calculate first sigma.sq since all the same

    rst = merge(rst, grpTab, by.all = "snp_id", all.x = T, all.y = F)
    rst = rst[order(rst$orgSeq),rstOutVal]

    #step2.  adding cov matrix
    genelist = unique(rst$gene_id)

    write.table(paste("#sigma.sq",sigma,sep = "="), file = ofile1, col.names = F, row.names = F,append = F, quote = F, sep = "\t")
    for(j in genelist){
        rstPec = rst[rst$gene_id == j,]
        snplist = rstPec$snp_id
        covPec = covNew[covNew$gene ==j,]
        size = dim(rstPec)[1]
        myMx = matrix(0, size,size)
        for(ii in 1:size){
            for(kk in ii:size){
                prt1 = covPec$val[covPec$snpA == snplist[ii] &
                                     covPec$snpB == snplist[kk]]
                if(length(prt1)==1){
                    myMx[ii,kk] = as.numeric(prt1)*sampleNo
                }
            }
        }
        write.table(cbind(rstPec, myMx), file = ofile1, col.names = F, row.names = F,append = T, quote = F, sep = "\t")
    }
}

##########################################
#.metaSKATToMass.  Preprocess metaData from MetaSKAT program
#input files:
#	ifile1: MSSD file name (full path)
#	ifile2: Minfo file name (full path)
#     ofile1: output meta data file name (full path)
#output:
##########################################
.metaSKATToMass = function(ifile1, ifile2, ofile1){
    #open input files using MetaSKAT::Open_MSSD_File_2Read()
    capture.output(assign("cohort", Open_MSSD_File_2Read(ifile1, ifile2)), file = ".sinkhole.txt")
    #cohort = Open_MSSD_File_2Read(ifile1, ifile2)
    sampleNo = cohort$EachInfo[[1]]$header$N.Sample
    srcTab = cohort$EachInfo[[1]]$Info
    geneList =  cohort$Set_unique

    unlink(ofile1)

    for(i in geneList){
        oneGene = MetaSKAT:::Get_META_Data_OneSet(cohort, i)
        rst = data.frame(gene_id = i,
                 snp_id = oneGene$Info.list[[1]]$SNPID,
                 mac = oneGene$Info.list[[1]]$MAF*2*sampleNo , 
                 N_obs = sampleNo,
                 score_stat = oneGene$Info.list[[1]]$Score,
                 stringsAsFactors = F)


        write.table(cbind(rst, oneGene$SMat.list[[1]]), file = ofile1, col.names = F, row.names = F,append = T, quote = F, sep = "\t")
    }
}

##########################################
#.seqMetaToMass.  Preprocess metaData from seqMata program
#input files:
#	ifile1: .RData file name(full path)
#     ofile1: output meta data file name (full path)
#     objName:  the obj name that have the data to be processed
#output:
##########################################
.seqMetaToMass = function(ifile1,ofile1, objName ){
    #open input files 
    load(ifile1)
    flag = 0
    if(objName %in% ls()){
        flag = 1
        tmpObj = get(objName)
        geneList = names(tmpObj)

        unlink(ofile1)
        for(i in geneList){
            oneGene = tmpObj[i][[1]]
            rst = data.frame(gene_id = i,
                 snp_id = names(oneGene$scores),
                 mac = oneGene$maf*2*oneGene$n , 
                 N_obs = oneGene$n,
                 score_stat = oneGene$scores/oneGene$sey^2,
                 stringsAsFactors = F)
            write.table(cbind(rst,as.matrix(oneGene$cov/oneGene$sey^2)), file = ofile1, col.names = F, row.names = F,append = T, quote = F, sep = "\t")
        }
    }
    flag
}

##########################################
#preMASS
#scriptFile, 
#outScript, 
#verson="1.0".  give out version number
##########################################
preMASS = function(scriptFile, outScript = paste(scriptFile, "out", sep = "."), version="1.0"){    
    #read in script file
    content = readLines(scriptFile)
    content = content[content != ""]
    content = content[!content %in% grep("^#",content, value = T)]
    start =  grep("^SOFTWARE", toupper(content))
    procNum = length(start)
    end = c(start[2:procNum]-1, length(content))

    #first loop check validation of the script
    errorCt = 0; #counting script file errors
    for(i in 1:procNum){
        spos = start[i]
        epos = end[i]
 
        software = toupper(gsub(" ", "", strsplit(content[spos], "=")[[1]][2]))

        if(software == "RAREMETAL"){
            FILE_SCORE = ""
            FILE_COV = ""
            FILE_GROUP = ""

            for(j in (spos+1):epos){
                oneItem = gsub("^[[:blank:]]*|[[:blank:]]*$", "", strsplit(content[j], "=")[[1]])
                if(toupper(oneItem[1]) == "FILE_SCORE"){
                    FILE_SCORE = oneItem[2]
                    if(!file.exists(FILE_SCORE)){
                        cat('"',FILE_SCORE,'"', " does not exist\n", sep = "")
                        errorCt = errorCt +1
                    }
                }

                if(toupper(oneItem[1]) == "FILE_COV"){
                    FILE_COV = oneItem[2]
                    if(!file.exists(FILE_COV)){
                        cat('"',FILE_COV, '"'," does not exist\n", sep = "")
                        errorCt = errorCt +1
                    }
                }

                if(toupper(oneItem[1]) == "FILE_GROUP"){
                    FILE_GROUP = oneItem[2]
                    if(!file.exists(FILE_GROUP)){
                        cat('"',FILE_GROUP, '"'," does not exist\n", sep = "")
                        errorCt = errorCt +1
                    }
                }
            }
        }else if(software == "METASKAT"){
            FILE_MSSD = ""
            FILE_MINFO = ""

            for(j in (spos+1):epos){
                oneItem = gsub("^ | $", "", strsplit(content[j], "=")[[1]])
                if(toupper(oneItem[1]) == "FILE_MSSD"){
                    FILE_MSSD = oneItem[2]
                    if(!file.exists(FILE_MSSD)){
                        cat('"',FILE_MSSD, '"'," does not exist.\n", sep = "")
                        errorCt = errorCt +1
                    }
                }

                if(toupper(oneItem[1]) == "FILE_MINFO"){
                    FILE_MINFO = oneItem[2]
                    if(!file.exists(FILE_MINFO)){
                        cat('"',FILE_MINFO,'"', " does not exist\n", sep = "")
                        errorCt = errorCt +1
                    }
                }
            }
        }else if(software == "SEQMETA"){
            FILE_RDATA = ""
            FILE_OBJ = ""

            for(j in (spos+1):epos){
                oneItem = gsub("^ | $", "", strsplit(content[j], "=")[[1]])
                if(toupper(oneItem[1]) == "FILE_RDATA"){
                    FILE_RDATA = oneItem[2]
                    if(!file.exists(FILE_RDATA)){
                        cat('"',FILE_RDATA, '"'," does not exist\n", sep = "")
                        errorCt = errorCt +1
                    }
                }

                if(toupper(oneItem[1]) == "FILE_OBJ"){
                    FILE_OBJ = oneItem[2]
                    if(FILE_OBJ == ""){
                        cat("FILE_OBJ must be specified\n")
                        errorCt = errorCt +1
                    }
                }
            }
        }else {
            #error condition stop processing
                cat("software type does not exist\n")
                errorCt = errorCt +1
        }

    }

    #second loop processes the data, if no data input error found
    if(errorCt > 0){
        cat(errorCt, " error(s) found.  Process terminated!  Please make change on your script and restart the process\n")

    }else{
        cat(procNum, "datasets to be processed...\n")
        unlink(outScript)
        stringX = "GENE_ID_COLUMN = 1\nGVAR_ID_COLUMN = 2\nMAC_COLUMN = 3\nN_OBS_COLUMN = 4\nSCORE_COLUMN = 5\n"

        for(i in 1:procNum){
            spos = start[i]
            epos = end[i]
            software = toupper(gsub(" ", "", strsplit(content[spos], "=")[[1]][2]))
 
            if(software == "RAREMETAL"){
                cat("\nStudy", i, ": RAREMETAL preprocessing\n")
                FILE_SCORE = ""
                FILE_COV = ""
                FILE_GROUP = ""

                for(j in (spos+1):epos){
                    oneItem = gsub("^ | $", "", strsplit(content[j], "=")[[1]])
                    if(toupper(oneItem[1]) == "FILE_SCORE"){
                        FILE_SCORE = oneItem[2]
                    }

                    if(toupper(oneItem[1]) == "FILE_COV"){
                        FILE_COV = oneItem[2]
                    }

                    if(toupper(oneItem[1]) == "FILE_GROUP"){
                        FILE_GROUP = oneItem[2]
                    }
                }
                temp = strsplit(FILE_SCORE, "/")[[1]]
                temp[length(temp)] = paste("MASS_", temp[length(temp)], ".txt", sep = "")
                ofile = paste0(temp, collapse = "/")
                .rareMetalToMass(FILE_SCORE, FILE_COV, FILE_GROUP, ofile); #processing
                stringY = paste("FILE = ", ofile,"\n", stringX,sep = "")
                write(stringY, file = outScript, append = T)
                
            }else if(software == "METASKAT"){
                cat("\nStudy", i, ": MetaSKAT preprocessing\n")
                FILE_MSSD = ""
                FILE_MINFO = ""

                for(j in (spos+1):epos){
                    oneItem = gsub("^ | $", "", strsplit(content[j], "=")[[1]])
                    if(toupper(oneItem[1]) == "FILE_MSSD"){
                        FILE_MSSD = oneItem[2]
                    }

                    if(toupper(oneItem[1]) == "FILE_MINFO"){
                        FILE_MINFO = oneItem[2]
                    }
                }

                temp = strsplit(FILE_MSSD, "/")[[1]]
                temp[length(temp)] = paste("MASS_", temp[length(temp)], ".txt", sep = "")
                ofile = paste0(temp, collapse = "/")

                .metaSKATToMass(FILE_MSSD, FILE_MINFO, ofile)
                stringY = paste("FILE = ", ofile,"\n", stringX,sep = "")
                write(stringY, file = outScript, append = T)
            }else if(software == "SEQMETA"){
                cat("\nStudy", i, ": seqMeta preprocessing\n")
                FILE_RDATA = ""
                FILE_OBJ = ""

                for(j in (spos+1):epos){
                    oneItem = gsub("^ | $", "", strsplit(content[j], "=")[[1]])
                    if(toupper(oneItem[1]) == "FILE_RDATA"){
                        FILE_RDATA = oneItem[2]
                    }

                    if(toupper(oneItem[1]) == "FILE_OBJ"){
                        FILE_OBJ = oneItem[2]
                    }
                }

                if(FILE_OBJ == ""){
                    temp = strsplit(FILE_RDATA, "/")[[1]]
                    FILE_OBJ = strsplit(temp[length(temp)], "\\.")[[1]][1]
                }
 
                temp = strsplit(FILE_RDATA, "/")[[1]]
                temp[length(temp)] = paste("MASS_", temp[length(temp)], ".txt", sep = "")
                ofile = paste0(temp, collapse = "/")

                flgs= .seqMetaToMass(FILE_RDATA, ofile, FILE_OBJ)
                if(flgs == 1){
                    stringY = paste("FILE = ", ofile,"\n", stringX,sep = "")
                    write(stringY, file = outScript, append = T)
                }else{
                    cat("Data file error detected, skip\n")
                }
            }
        };#end of for/i loop
    }; #end of if::errorCt statement
    unlink(".sinkhole.txt")
}


###########################################
#premeta
#
##########################################
PreMeta = function(scriptFile, software, version = "1.0"){
    if(software == "MASS"){
        preMASS(scriptFile,version = version) 
    }else if(software == "seqMeta"){
        cat("UNDER DEVELOPMENT")
    }else if(software == "RAREMETAL"){
        cat("UNDER DEVELOPMENT")
    }else if(software == "MetaSKAT"){
        cat("UNDER DEVELOPMENT")
    }else{
        cat("Invalid output software type.")
    }
}




